from __future__ import annotations

import json
from functools import cached_property
from os.path import basename
from pathlib import Path

from pydantic import ByteSize, field_validator
from torf import Torrent as TorfTorrent

from seadex._models import FrozenBaseModel
from seadex._types import StrPath
from seadex._utils import realpath


class File(FrozenBaseModel):
    """Represents a file in the torrent."""

    name: str
    """The name of the file."""
    size: ByteSize
    """The size of the file in bytes."""

    def __str__(self) -> str:
        """Stringify, equivalent to [`File.name`][File.name]."""
        return self.name

    def __fspath__(self) -> str:
        """Path representation, equivalent to [`File.name`][File.name]."""
        return self.name

    @field_validator("name", mode="before")
    @classmethod
    def _as_posix(cls, v: str) -> str:
        """Ensure the names are posix compatible."""
        return Path(v).as_posix()


class FileList(tuple[File, ...]):
    """A tuple-based collection of `File` objects."""

    def to_json(self, indent: int | None = None) -> str:
        """
        Convert the file list to a JSON string that's compatible with SeaDex.

        Parameters
        ----------
        indent : int | None, optional
            Number of spaces for indentation in the output JSON string.

        Returns
        -------
        str
            The file list in JSON format.

        """
        files = []

        for file in self:
            files.append({"filename": basename(file.name), "size": int(file.size)})
        return json.dumps(files, indent=indent)


class SeaDexTorrent:
    def __init__(self, file: StrPath) -> None:
        """
        Class to handle torrent files for SeaDex.

        Parameters
        ----------
        file : StrPath
            The path to the torrent file.

        """
        self._file = realpath(file)

    @property
    def file(self) -> Path:
        """Resolved path to the torrent file."""
        return self._file

    @cached_property
    def filelist(self) -> FileList:
        """List of files within the torrent."""
        torrent = TorfTorrent.read(self.file)
        files = []
        for file in torrent.files:
            files.append(File(name=file, size=file.size))
        return FileList(files)

    def sanitize(self, *, destination: StrPath | None = None, overwrite: bool = False) -> Path:
        """
        Sanitizes the torrent file by removing sensitive data and optionally saves it to a new location.

        Parameters
        ----------
        destination : StrPath | None, optional
            The destination path to save the sanitized torrent. If None, the sanitized file is saved in place.
        overwrite : bool, optional
            If True, overwrites the existing file or destination file if it exists.

        Returns
        -------
        Path
            The path to the sanitized torrent file.

        Raises
        ------
        FileExistsError
            - If `destination` is None and `overwrite` is False.
            - If `destination` already exists and `overwrite` is False.

        Notes
        -----
        - If the torrent file is public (i.e., not marked as private), it is returned as is.
        - The following fields are removed from the torrent file if it is private:
            - Trackers
            - Web seeds
            - HTTP seeds
            - Private flag
            - Comment
            - Creation date
            - Created by field
            - Source field
        - The torrent's `infohash` is randomized.

        """
        torrent = TorfTorrent.read(self.file)

        if not torrent.private:
            # Public torrent
            return self.file

        torrent.trackers = None
        torrent.webseeds = None
        torrent.httpseeds = None
        torrent.private = None
        torrent.comment = None
        torrent.creation_date = None
        torrent.created_by = None
        torrent.source = None
        torrent.randomize_infohash = True

        if destination is None:
            if overwrite is False:
                raise FileExistsError(f"{self.file} already exists!")
            else:
                torrent.write(self.file, overwrite=True)
                return self.file
        else:
            destination = realpath(destination)
            torrent.write(destination, overwrite=overwrite)
            return destination



from __future__ import annotations

from datetime import datetime, timezone
from os import PathLike
from typing import Annotated, TypeAlias

from pydantic import AfterValidator

StrPath: TypeAlias = str | PathLike[str]
"""String or path-like objects"""

UTCDateTime: TypeAlias = Annotated[datetime, AfterValidator(lambda dt: dt.astimezone(timezone.utc))]
"""Datetime that's always in UTC."""



from __future__ import annotations

import re
import shutil
from datetime import datetime, timezone
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Any
from urllib.parse import urljoin
from uuid import uuid4
from zipfile import ZipFile

from httpx import Client
from pydantic import ByteSize
from typing_extensions import assert_never

from seadex._exceptions import BadBackupFileError
from seadex._models import FrozenBaseModel
from seadex._types import StrPath, UTCDateTime
from seadex._utils import httpx_client, realpath

if TYPE_CHECKING:
    from typing_extensions import Self


class BackupFile(FrozenBaseModel):
    """A model representing a backup file."""

    name: str
    """The name of the backup file."""
    size: ByteSize
    """The size of the backup file in bytes."""
    modified_time: UTCDateTime
    """The last modified time of the backup file."""

    def __str__(self) -> str:
        """Implement the string representation. Equivalent to `BackupFile.name`."""
        return self.name

    def __fspath__(self) -> str:
        """
        Path representation. Equivalent to `BackupFile.name`.
        Allows for compatibility with `PathLike` objects.

        Examples
        --------
        >>> from pathlib import Path
        >>> from seadex import BackupFile
        >>> backup = BackupFile(name="20240909041339-seadex-backup.zip", size=..., modified=...)
        >>> Path.home() / backup
        PosixPath('/home/raven/20240909041339-seadex-backup.zip')

        """
        return self.name

    @classmethod
    def _from_dict(cls, dictionary: dict[str, Any], /) -> Self:
        """Parse the response from the SeaDex Backup API into a `BackupFile` object."""
        kwargs = {
            "name": dictionary["key"],
            "modified_time": dictionary["modified"],
            "size": dictionary["size"],
        }
        return cls.model_validate(kwargs)


class SeaDexBackup:
    def __init__(
        self, email: str, password: str, base_url: str = "https://releases.moe", client: Client | None = None
    ) -> None:
        """
        Client to interact with the SeaDex backup API.

        Parameters
        ----------
        email : str
            The email address for authentication.
        password : str
            The password for authentication.
        base_url : str, optional
            The base URL of SeaDex, used for constructing API queries.
        client : Client, optional
            An [`httpx.Client`](https://www.python-httpx.org/api/#client) instance used to make requests to SeaDex.

        Examples
        --------
        ```py
        with SeaDexBackup("me@email.com", "password") as seadex_backup:
            print(seadex_backup.latest_backup)
            #> @auto_pb_backup_sea_dex_20241122000000.zip
        ```

        Notes
        -----
        Only SeaDex admins can use this! Logging in with a non-admin account will result in failure.

        """
        self._base_url = base_url
        self._client = httpx_client() if client is None else client
        self._admin_token = self._auth_with_password(email, password)

    @property
    def base_url(self) -> str:
        """
        Base URL, used for constructing API queries.
        """
        return self._base_url

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args: object) -> None:
        self.close()

    def close(self) -> None:
        """
        Close the underlying HTTP client connection.
        """
        self._client.close()

    def _url_for(self, endpoint: str) -> str:
        return urljoin(self._base_url, endpoint)

    def _auth_with_password(self, email: str, password: str) -> str:
        response = self._client.post(
            self._url_for("/api/admins/auth-with-password"), json={"identity": email, "password": password}
        )
        admin = response.raise_for_status().json()
        return admin["token"]  # type: ignore[no-any-return]

    def _get_file_token(self) -> str:
        response = self._client.post(self._url_for("/api/files/token"), headers={"Authorization": self._admin_token})
        return response.raise_for_status().json()["token"]  # type: ignore[no-any-return]

    @property
    def backups(self) -> tuple[BackupFile, ...]:
        """
        Retrieve a tuple of backup files.

        Returns
        -------
        tuple[BackupFile, ...]
            A tuple of backup files, sorted by the modified date.

        """

        response = self._client.get(
            "https://releases.moe/api/backups", headers={"Authorization": self._admin_token}
        ).json()
        backups = (BackupFile._from_dict(backup) for backup in response)
        return tuple(sorted(backups, key=lambda f: f.modified_time))

    @property
    def latest_backup(self) -> BackupFile:
        """
        Retrieve the latest backup file.

        Returns
        -------
        BackupFile
            The latest backup file.

        """
        return self.backups[-1]

    def download(self, file: str | BackupFile | None = None, *, destination: StrPath = Path.cwd()) -> Path:
        """
        Download the specified backup file to the given destination directory.

        Parameters
        ----------
        file : str | BackupFile | None, optional
            The backup file to download. If `None`, downloads the [latest existing backup][seadex.SeaDexBackup.latest_backup].
        destination : StrPath, optional
            The destination directory to save the backup.

        Returns
        -------
        Path
            The path to the downloaded backup file.

        Raises
        ------
        NotADirectoryError
            If the destination is not a valid directory.
        BadBackupFileError
            if the downloaded backup file fails integrity check.

        """
        destination = realpath(destination)

        if not destination.is_dir():
            raise NotADirectoryError(f"{destination} must be an existing directory!")

        match file:
            case None:
                key = self.latest_backup.name
            case str():
                key = file
            case BackupFile():
                key = file.name
            case _:
                assert_never(file)

        outfile = destination / key

        with TemporaryDirectory(ignore_cleanup_errors=True) as tmpdir:
            tmpfile = Path(tmpdir).resolve() / str(uuid4())
            response = self._client.get(self._url_for(f"/api/backups/{key}"), params={"token": self._get_file_token()})
            tmpfile.write_bytes(response.content)
            shutil.move(tmpfile, outfile)

        with ZipFile(outfile) as archive:
            check = archive.testzip()

        if check is not None:
            outfile.unlink(missing_ok=True)
            raise BadBackupFileError(f"{outfile} failed integrity check!")

        return outfile

    def create(self, filename: str) -> BackupFile:
        """
        Create a new backup with the specified filename.

        Parameters
        ----------
        filename : str
            The name to assign to the backup file.
            This can include full formatting options as supported by
            [`datetime.strftime`][datetime.datetime.strftime].

        Returns
        -------
        BackupFile
            The newly created backup file.

        """
        _filename = filename.removesuffix(".zip") + ".zip"
        _filename = datetime.now(timezone.utc).strftime(_filename).casefold()

        if re.match(r"^([a-z0-9_-]+\.zip)$", _filename) is None:
            # The API forbids anything else, so we need to enforce it.
            raise ValueError(
                f"Invalid filename: {_filename!r}. The filename may only contain alphanumeric characters, hyphens, or underscores."
            )

        self._client.post(
            self._url_for("/api/backups"), json={"name": _filename}, headers={"Authorization": self._admin_token}
        ).raise_for_status()

        return next(filter(lambda member: member.name == _filename, self.backups))

    def delete(self, file: str | BackupFile) -> None:
        """
        Delete the specified backup file.

        Parameters
        ----------
        file : str | BackupFile
            The backup file to delete.

        Returns
        -------
        None

        """
        self._client.delete(
            self._url_for(f"/api/backups/{file}"), headers={"Authorization": self._admin_token}
        ).raise_for_status()



from __future__ import annotations

import pytest
from pytest_httpx import HTTPXMock

from seadex import EntryNotFoundError, SeaDexEntry


def test_entry_not_found_from_anilist_id(seadex_entry: SeaDexEntry, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://releases.moe/api/collections/entries/records?filter=alID=98329837198378237183918&expand=trs",
        json={"page": 1, "perPage": 30, "totalItems": 0, "totalPages": 0, "items": []},
    )

    with pytest.raises(EntryNotFoundError):
        seadex_entry.from_id(98329837198378237183918)


def test_entry_not_found_from_seadex_id(seadex_entry: SeaDexEntry, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://releases.moe/api/collections/entries/records?filter=id='jashdsjhdsakjdhsakdhsadhaksjd'&expand=trs",
        json={"page": 1, "perPage": 30, "totalItems": 0, "totalPages": 0, "items": []},
    )

    with pytest.raises(EntryNotFoundError):
        seadex_entry.from_id("jashdsjhdsakjdhsakdhsadhaksjd")


def test_entry_not_found_from_title(seadex_entry: SeaDexEntry, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://graphql.anilist.co",
        json={
            "errors": [{"message": "Not Found.", "status": 404, "locations": [{"line": 1, "column": 28}]}],
            "data": {"Media": None},
        },
    )

    with pytest.raises(EntryNotFoundError):
        seadex_entry.from_title("jashdsjhdsakjdhsakdhsadhaksjd")



from __future__ import annotations

from pydantic import BaseModel, ConfigDict


class FrozenBaseModel(BaseModel):
    """Frozen pydantic.BaseModel."""

    model_config = ConfigDict(frozen=True, extra="forbid")



from __future__ import annotations

from os.path import basename
from typing import TYPE_CHECKING
from urllib.parse import urljoin

from seadex._exceptions import EntryNotFoundError
from seadex._records import EntryRecord
from seadex._types import StrPath
from seadex._utils import httpx_client

if TYPE_CHECKING:
    from collections.abc import Iterator

    from httpx import Client
    from typing_extensions import Self


class SeaDexEntry:
    def __init__(self, base_url: str = "https://releases.moe", client: Client | None = None) -> None:
        """
        Client to interact with the anime entries in SeaDex.

        Parameters
        ----------
        base_url : str, optional
            The base URL of SeaDex, used for constructing API queries.
        client : Client, optional
            An [`httpx.Client`](https://www.python-httpx.org/api/#client) instance used to make requests to SeaDex.

        Examples
        --------
        ```py
        with SeaDexEntry() as entry:
            tamako = entry.from_title("tamako love story")
            for torrent in tamako.torrents:
                if torrent.is_best and torrent.tracker.is_public():
                    print(torrent.release_group)
                    #> LYS1TH3A
                    #> Okay-Subs
        ```

        """
        self._base_url = base_url
        self._endpoint = urljoin(self._base_url, "/api/collections/entries/records")
        self._client = httpx_client() if client is None else client

    @property
    def base_url(self) -> str:
        """
        Base URL, used for constructing API queries.
        """
        return self._base_url

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args: object) -> None:
        self.close()

    def close(self) -> None:
        """
        Close the underlying HTTP client connection.
        """
        self._client.close()

    def from_id(self, id: int | str, /) -> EntryRecord:
        """
        Retrieve an entry by its ID.

        Parameters
        ----------
        id : int | str
            The ID of the entry. Can be an AniList ID (integer)
            or a SeaDex database ID (string).

        Returns
        -------
        EntryRecord
            The retrieved entry.

        Raises
        ------
        EntryNotFoundError
            If no entry is found for the provided ID.

        """

        if isinstance(id, int):
            params = {"filter": f"alID={id}", "expand": "trs"}  # Anilist IDs are integers
        else:
            params = {"filter": f"id='{id}'", "expand": "trs"}  # Database IDs are strings

        response = self._client.get(self._endpoint, params=params).raise_for_status()

        try:
            return EntryRecord._from_dict(response.json()["items"][0])
        except (KeyError, IndexError):
            errmsg = f"No seadex entry found for id: {id}"
            raise EntryNotFoundError(errmsg)

    def from_title(self, title: str, /) -> EntryRecord:
        """
        Retrieve an entry by its anime title.

        Parameters
        ----------
        title : str
            The title of the anime to search for.

        Returns
        -------
        EntryRecord
            The retrieved entry.

        Raises
        ------
        EntryNotFoundError
            If no entry is found for the provided title.

        """
        try:
            response = self._client.post(
                "https://graphql.anilist.co",
                json={
                    "query": "query ($search: String!) { Media(search: $search, type: ANIME) { id title { english romaji } } }",
                    "variables": {"search": title},
                },
            ).raise_for_status()

            media = response.json()["data"]["Media"]
            anilist_id = media["id"]

            response = self._client.get(
                self._endpoint, params={"filter": f"alID={anilist_id}", "expand": "trs"}
            ).raise_for_status()

            entry_record = EntryRecord._from_dict(response.json()["items"][0])
            entry_record._anilist_title = media["title"]["english"] or media["title"]["romaji"]  # type: ignore[attr-defined]
            return entry_record

        except (KeyError, IndexError, TypeError):
            errmsg = f"No seadex entry found for title: {title}"
            raise EntryNotFoundError(errmsg)

    def from_filename(self, filename: StrPath, /) -> Iterator[EntryRecord]:
        """
        Yield entries that may contain a file with the specified filename.

        Parameters
        ----------
        filename : StrPath
            The filename to search for.

        Yields
        ------
        EntryRecord
            The retrieved entry.

        """
        basefilename = basename(filename)
        params = {"filter": f'trs.files?~\'"name":"{basefilename}"\'', "expand": "trs"}
        response = self._client.get(self._endpoint, params=params).raise_for_status()
        for entry in response.json()["items"]:
            yield EntryRecord._from_dict(entry)

    def iterator(self) -> Iterator[EntryRecord]:
        """
        Lazily iterate over all the entries in SeaDex.

        Yields
        ------
        EntryRecord
            The retrieved entry.

        """
        total_pages = (
            self._client.get(self._endpoint, params={"perPage": 500}).raise_for_status().json()["totalPages"] + 1
        )

        for page in range(1, total_pages):
            response = self._client.get(
                self._endpoint, params={"page": page, "perPage": 500, "expand": "trs"}
            ).raise_for_status()
            for item in response.json()["items"]:
                yield EntryRecord._from_dict(item)



from __future__ import annotations

import pytest

from seadex import Tracker


@pytest.mark.parametrize(
    "tracker, value",
    [
        (Tracker.NYAA, "Nyaa"),
        (Tracker.ANIMETOSHO, "AnimeTosho"),
        (Tracker.ANIDEX, "AniDex"),
        (Tracker.RUTRACKER, "RuTracker"),
        (Tracker.PRIVATE_TRACKER, "PT"),
        (Tracker.BEYONDHD, "BeyondHD"),
        (Tracker.PASSTHEPOPCORN, "PassThePopcorn"),
        (Tracker.BROADCASTTHENET, "BroadcastTheNet"),
        (Tracker.HDBITS, "HDBits"),
        (Tracker.BLUTOPIA, "Blutopia"),
        (Tracker.AITHER, "Aither"),
        (Tracker.OTHER, "Other"),
    ],
)
def test_tracker_values(tracker: Tracker, value: str) -> None:
    assert tracker == value


@pytest.mark.parametrize(
    "tracker, is_private, is_public",
    [
        ("Nyaa", False, True),
        ("AnimeTosho", False, True),
        ("AniDex", False, True),
        ("RuTracker", False, True),
        ("PT", True, False),
        ("BeyondHD", True, False),
        ("PassThePopcorn", True, False),
        ("BroadcastTheNet", True, False),
        ("HDBits", True, False),
        ("Blutopia", True, False),
        ("Aither", True, False),
        ("Other", True, False),
    ],
)
def test_tracker_is_private(tracker: str, is_private: bool, is_public: bool) -> None:
    assert Tracker(tracker).is_private() == is_private
    assert Tracker(tracker).is_public() == is_public


@pytest.mark.parametrize(
    "tracker, domain",
    [
        ("Nyaa", "nyaa.si"),
        ("AnimeTosho", "animetosho.org"),
        ("AniDex", "anidex.info"),
        ("RuTracker", "rutracker.org"),
        ("BeyondHD", "beyond-hd.me"),
        ("PassThePopcorn", "passthepopcorn.me"),
        ("BroadcastTheNet", "broadcasthe.net"),
        ("HDBits", "hdbits.org"),
        ("Blutopia", "blutopia.cc"),
        ("Aither", "aither.cc"),
        ("Other", ""),
    ],
)
def test_tracker_domain(tracker: str, domain: str) -> None:
    assert Tracker(tracker).domain == domain


def test_bad_value() -> None:
    with pytest.raises(ValueError):
        Tracker("kasdjsahdjshdahdakjds")

    with pytest.raises(ValueError):
        Tracker(121212)


def test_case_insensitive_lookup() -> None:
    assert Tracker("nyAA") is Tracker.NYAA
    assert Tracker("ANIMETOSHO") is Tracker.ANIMETOSHO
    assert Tracker("pt") is Tracker.PRIVATE_TRACKER



from __future__ import annotations

from cyclopts import App
from rich import box, print_json
from rich.console import Console, Group
from rich.table import Table

from seadex._entry import SeaDexEntry
from seadex._exceptions import EntryNotFoundError
from seadex._version import __version__

entry_app = App(
    "entry",
    version=__version__,
    help="Get a SeaDex entry.",
    help_format="plaintext",
    version_flags=None,
)


@entry_app.default
def get_entry(title: str, /, *, json: bool = False, pretty: bool = False) -> None:
    """
    Get the seadex entry for the given title.

    Parameters
    ----------
    title : str
        Title of the anime.
    json : bool, optional
        If True, the output will be in JSON.
    pretty : bool, optional
        If True, the JSON output will be pretty-printed.

    """
    console = Console()

    with SeaDexEntry() as seadex_entry:
        try:
            entry = seadex_entry.from_title(title)
        except EntryNotFoundError as error:
            console.print(f"[red]error:[/] {error.message}")
            return

        if json:
            if pretty:
                print_json(entry.model_dump_json())
                return
            else:
                print(entry.model_dump_json())
                return

        body = f"Title: {entry._anilist_title}\n"  # type: ignore[attr-defined]
        body += f"URL: {entry.url}\n"
        body += f"AniList: https://anilist.co/anime/{entry.anilist_id}\n"
        body += f"Incomplete: {entry.is_incomplete}\n"
        body += f"Updated At: {entry.updated_at.strftime('%Y-%m-%d')}"
        if entry.theoretical_best is not None:
            body += f"\nTheoretical Best: {entry.theoretical_best}"

        table = Table(box=box.MARKDOWN)
        table.add_column("Group")
        table.add_column("Best")
        table.add_column("Dual")
        table.add_column("URL")

        for torrent in sorted(
            entry.torrents, key=lambda t: (not (t.is_best and t.is_dual_audio), not t.is_best, not t.is_dual_audio)
        ):
            table.add_row(
                torrent.release_group,
                ":white_check_mark:" if torrent.is_best else ":cross_mark:",
                ":white_check_mark:" if torrent.is_dual_audio else ":cross_mark:",
                torrent.url,
            )

        lower_body = f"Notes:\n{entry.notes}\n"
        lower_body += "\nComparisons:\n"
        for comparison in entry.comparisons:
            lower_body += f"- {comparison}\n"

        console.print(Group(body, table, lower_body))



from __future__ import annotations

from datetime import datetime, timezone

from pytest_httpx import HTTPXMock

from seadex import SeaDexEntry, Tracker

SAMPLE_JSON_REPLY = {
    "page": 1,
    "perPage": 30,
    "totalItems": 1,
    "totalPages": 1,
    "items": [
        {
            "alID": 20519,
            "collectionId": "3l2x9nxip35gqb5",
            "collectionName": "entries",
            "comparison": "https://slow.pics/c/rc6qrB1F",
            "created": "2024-01-30 19:28:10.337Z",
            "expand": {
                "trs": [
                    {
                        "collectionId": "oiwizhmushn5qqh",
                        "collectionName": "torrents",
                        "created": "2024-01-30 19:28:09.110Z",
                        "dualAudio": True,
                        "files": [
                            {
                                "length": 4636316199,
                                "name": "Tamako.Love.Story.2014.1080p.BluRay.Opus2.0.H.265-LYS1TH3A.mkv",
                            }
                        ],
                        "id": "pcpina3ekbqk7a5",
                        "infoHash": "23f77120cfdf9df8b42a10216aa33e281c58b456",
                        "isBest": True,
                        "releaseGroup": "LYS1TH3A",
                        "tracker": "Nyaa",
                        "updated": "2024-01-30 19:28:09.110Z",
                        "url": "https://nyaa.si/view/1693872",
                    },
                    {
                        "collectionId": "oiwizhmushn5qqh",
                        "collectionName": "torrents",
                        "created": "2024-01-30 19:28:09.461Z",
                        "dualAudio": True,
                        "files": [
                            {
                                "length": 4636316199,
                                "name": "Tamako.Love.Story.2014.1080p.BluRay.Opus2.0.H.265-LYS1TH3A.mkv",
                            }
                        ],
                        "id": "tvh4fn4m2qi19n5",
                        "infoHash": "<redacted>",
                        "isBest": True,
                        "releaseGroup": "LYS1TH3A",
                        "tracker": "PT",
                        "updated": "2024-01-30 19:28:09.461Z",
                        "url": "/torrents.php?id=20684&torrentid=1053072",
                    },
                ]
            },
            "id": "c344w8ld7q1yppz",
            "incomplete": False,
            "notes": "Okay-Subs is JPN BD Encode+Commie with additional honorifics track\nLYS1TH3A is Okay-Subs+Dub",
            "theoreticalBest": "",
            "trs": ["pcpina3ekbqk7a5", "tvh4fn4m2qi19n5", "qhcmujh4dsw55j2", "enytf1g1cxf0k47"],
            "updated": "2024-01-30 19:28:10.337Z",
        }
    ],
}


def test_properties(seadex_entry: SeaDexEntry) -> None:
    assert seadex_entry.base_url == "https://releases.moe"


def test_from_anilist_id(seadex_entry: SeaDexEntry, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://releases.moe/api/collections/entries/records?filter=alID%3D20519&expand=trs",
        json=SAMPLE_JSON_REPLY,
    )
    entry = seadex_entry.from_id(20519)
    assert entry.anilist_id == 20519
    assert entry.collection_id == "3l2x9nxip35gqb5"
    assert entry.collection_name == "entries"
    assert entry.comparisons == ("https://slow.pics/c/rc6qrB1F",)
    assert entry.created_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)
    assert entry.id == "c344w8ld7q1yppz"
    assert not entry.is_incomplete
    assert (
        entry.notes == "Okay-Subs is JPN BD Encode+Commie with additional honorifics track\nLYS1TH3A is Okay-Subs+Dub"
    )
    assert entry.theoretical_best is None
    assert entry.torrents[0].url == "https://nyaa.si/view/1693872"
    assert entry.torrents[1].url == "/torrents.php?id=20684&torrentid=1053072"
    assert entry.torrents[0].infohash is not None
    assert entry.torrents[1].infohash is None
    assert entry.torrents[0].tracker == Tracker.NYAA
    assert entry.torrents[1].tracker == Tracker.PRIVATE_TRACKER
    assert entry.torrents[0].infohash is not None
    assert entry.torrents[1].infohash is None
    assert isinstance(entry.updated_at, datetime)


def test_from_seadex_id(seadex_entry: SeaDexEntry, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://releases.moe/api/collections/entries/records?filter=id='c344w8ld7q1yppz'&expand=trs",
        json=SAMPLE_JSON_REPLY,
    )
    entry = seadex_entry.from_id("c344w8ld7q1yppz")
    assert entry.anilist_id == 20519
    assert entry.collection_id == "3l2x9nxip35gqb5"
    assert entry.collection_name == "entries"
    assert entry.comparisons == ("https://slow.pics/c/rc6qrB1F",)
    assert entry.created_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)
    assert entry.id == "c344w8ld7q1yppz"
    assert not entry.is_incomplete
    assert (
        entry.notes == "Okay-Subs is JPN BD Encode+Commie with additional honorifics track\nLYS1TH3A is Okay-Subs+Dub"
    )
    assert entry.theoretical_best is None
    assert entry.torrents[0].url == "https://nyaa.si/view/1693872"
    assert entry.torrents[1].url == "/torrents.php?id=20684&torrentid=1053072"
    assert entry.torrents[0].infohash is not None
    assert entry.torrents[1].infohash is None
    assert entry.torrents[0].tracker == Tracker.NYAA
    assert entry.torrents[1].tracker == Tracker.PRIVATE_TRACKER
    assert entry.torrents[0].infohash is not None
    assert entry.torrents[1].infohash is None
    assert entry.updated_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)


def test_from_title(seadex_entry: SeaDexEntry, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://graphql.anilist.co",
        json={
            "data": {"Media": {"id": 20519, "title": {"english": "Tamako -love story-", "romaji": "Tamako Love Story"}}}
        },
    )

    httpx_mock.add_response(
        url="https://releases.moe/api/collections/entries/records?filter=alID%3D20519&expand=trs",
        json=SAMPLE_JSON_REPLY,
    )

    entry = seadex_entry.from_title("tamako love story")
    assert entry.anilist_id == 20519
    assert entry._anilist_title == "Tamako -love story-"  # type: ignore[attr-defined]
    assert entry.collection_id == "3l2x9nxip35gqb5"
    assert entry.collection_name == "entries"
    assert entry.comparisons == ("https://slow.pics/c/rc6qrB1F",)
    assert entry.created_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)
    assert entry.id == "c344w8ld7q1yppz"
    assert not entry.is_incomplete
    assert (
        entry.notes == "Okay-Subs is JPN BD Encode+Commie with additional honorifics track\nLYS1TH3A is Okay-Subs+Dub"
    )
    assert entry.theoretical_best is None
    assert entry.torrents[0].infohash is not None
    assert entry.torrents[1].infohash is None
    assert entry.torrents[0].tracker == Tracker.NYAA
    assert entry.torrents[1].tracker == Tracker.PRIVATE_TRACKER
    assert entry.torrents[0].infohash is not None
    assert entry.torrents[1].infohash is None
    assert entry.updated_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)


def test_from_filename(seadex_entry: SeaDexEntry, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://releases.moe/api/collections/entries/records?filter=trs.files%3F~%27%22name%22%3A%22Tamako.Love.Story.2014.1080p.BluRay.Opus2.0.H.265-LYS1TH3A.mkv%22%27&expand=trs",
        json=SAMPLE_JSON_REPLY,
    )

    entries = seadex_entry.from_filename("Tamako.Love.Story.2014.1080p.BluRay.Opus2.0.H.265-LYS1TH3A.mkv")
    entry = tuple(entries)[0]  # noqa: RUF015
    assert entry.anilist_id == 20519
    assert entry.collection_id == "3l2x9nxip35gqb5"
    assert entry.collection_name == "entries"
    assert entry.comparisons == ("https://slow.pics/c/rc6qrB1F",)
    assert entry.created_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)
    assert entry.id == "c344w8ld7q1yppz"
    assert not entry.is_incomplete
    assert (
        entry.notes == "Okay-Subs is JPN BD Encode+Commie with additional honorifics track\nLYS1TH3A is Okay-Subs+Dub"
    )
    assert entry.theoretical_best is None
    assert entry.torrents[0].infohash is not None
    assert entry.torrents[1].infohash is None
    assert entry.torrents[0].tracker == Tracker.NYAA
    assert entry.torrents[1].tracker == Tracker.PRIVATE_TRACKER
    assert entry.torrents[0].infohash is not None
    assert entry.torrents[1].infohash is None
    assert entry.updated_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)


def test_iterator(seadex_entry: SeaDexEntry, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://releases.moe/api/collections/entries/records?perPage=500",
        json={"totalPages": 1},
    )

    httpx_mock.add_response(
        url="https://releases.moe/api/collections/entries/records?page=1&perPage=500&expand=trs",
        json=SAMPLE_JSON_REPLY,
    )

    for entry in seadex_entry.iterator():
        assert entry.anilist_id == 20519
        assert entry.collection_id == "3l2x9nxip35gqb5"
        assert entry.collection_name == "entries"
        assert entry.comparisons == ("https://slow.pics/c/rc6qrB1F",)
        assert entry.created_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)
        assert entry.id == "c344w8ld7q1yppz"
        assert not entry.is_incomplete
        assert (
            entry.notes
            == "Okay-Subs is JPN BD Encode+Commie with additional honorifics track\nLYS1TH3A is Okay-Subs+Dub"
        )
        assert entry.theoretical_best is None
        assert entry.torrents[0].url == "https://nyaa.si/view/1693872"
        assert entry.torrents[1].url == "/torrents.php?id=20684&torrentid=1053072"
        assert entry.torrents[0].infohash is not None
        assert entry.torrents[1].infohash is None
        assert entry.torrents[0].tracker == Tracker.NYAA
        assert entry.torrents[1].tracker == Tracker.PRIVATE_TRACKER
        assert entry.torrents[0].infohash is not None
        assert entry.torrents[1].infohash is None
        assert entry.updated_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)



from __future__ import annotations

from cyclopts import App
from cyclopts.types import ResolvedExistingPath, ResolvedPath
from natsort import natsorted, ns
from rich import print_json
from rich.console import Console
from rich.progress import track

from seadex._torrent import SeaDexTorrent
from seadex._version import __version__

torrent_app = App(
    "torrent",
    version=__version__,
    help="Perform torrent operations.",
    help_format="plaintext",
    version_flags=None,
)


@torrent_app.command
def sanitize(src: ResolvedExistingPath, dst: ResolvedPath | None = None, /) -> None:
    """
    Sanitize torrent files by removing sensitive data.

    Parameters
    ----------
    src : ResolvedExistingPath
        Path to the source torrent file or directory containing torrent files to sanitize.
    dst : ResolvedPath or None, optional
        Path to the destination where sanitized files will be stored.

    """
    console = Console()

    if src.is_file():
        path = SeaDexTorrent(src).sanitize(destination=dst, overwrite=True)
        console.print(f":white_check_mark: Saved sanitized torrent to [cyan]{path}[/cyan]", emoji=True)
    else:
        if dst is None:
            console.print("[red]error:[/] destination must be an existing directory.")
            return

        if not dst.is_dir():
            console.print(f"[red]error:[/] {dst} must be an existing directory.")
            return

        files = natsorted(src.rglob("*.torrent"), alg=ns.PATH)
        for file in track(files, description="Sanitizing...", total=len(files), transient=True, console=console):
            path = SeaDexTorrent(file).sanitize(destination=dst / file.name, overwrite=True)
            console.print(f":white_check_mark: Saved sanitized torrent to [cyan]{path}[/]", emoji=True)


@torrent_app.command
def json(src: ResolvedExistingPath, /, *, pretty: bool = False, copy: bool = True) -> None:
    """
    Output the list of files in a torrent as a SeaDex compatible JSON string.

    Parameters
    ----------
    src : ResolvedExistingPath
        Path to the torrent file.
    pretty : bool, optional
        If True, the JSON output will be pretty-printed.
    copy : bool, optional
        Copy the JSON output to clipboard.

    """
    filelist = SeaDexTorrent(src).filelist

    if copy:  # pragma: no cover
        try:
            import pyperclip

            pyperclip.copy(filelist.to_json())
        except Exception:
            pass

    if pretty:
        print_json(filelist.to_json())
    else:
        print(filelist.to_json())



from __future__ import annotations

from pathlib import Path

from httpx import Client

from seadex._types import StrPath
from seadex._version import __version__


def realpath(path: StrPath) -> Path:
    """
    Resolve Path or Path-like strings and return a Path object.
    """
    return Path(path).expanduser().resolve()


def httpx_client() -> Client:
    """
    Return an instance of an httpx.Client.
    """
    return Client(headers={"user-agent": f"seadex/{__version__}"})



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from natsort import natsorted, ns
from pydantic import field_validator

from seadex._enums import Tracker
from seadex._models import FrozenBaseModel
from seadex._torrent import File
from seadex._types import UTCDateTime

if TYPE_CHECKING:
    from typing_extensions import Self


class TorrentRecord(FrozenBaseModel):
    """Represents a single torrent record within a SeaDex entry."""

    collection_id: str
    """The ID of the collection the torrent record belongs to."""
    collection_name: str
    """The name of the collection the torrent record belongs to."""
    created_at: UTCDateTime
    """The timestamp of when the torrent record was created."""
    is_dual_audio: bool
    """Whether the torrent contains both Japanese and English audio tracks."""
    files: tuple[File, ...]
    """A tuple of `File` objects representing the files in the torrent."""
    id: str
    """The ID of the torrent record."""
    infohash: str | None
    """The infohash of the torrent if available, otherwise `None` (private torrents)."""
    is_best: bool
    """Whether this torrent is marked as the "best"."""
    release_group: str
    """The name of the group that released the torrent."""
    tracker: Tracker
    """The tracker where the torrent is hosted."""
    updated_at: UTCDateTime
    """The timestamp of when the torrent record was last updated."""
    url: str
    """The URL of the torrent."""

    @classmethod
    def _from_dict(cls, dictionary: dict[str, Any], /) -> Self:
        """Parse the response from the SeaDex API into a `TorrentRecord` object."""
        kwargs = {
            "collection_id": dictionary["collectionId"],
            "collection_name": dictionary["collectionName"],
            "created_at": dictionary["created"],
            "is_dual_audio": dictionary["dualAudio"],
            "files": (dict(name=file["name"], size=file["length"]) for file in dictionary["files"]),
            "id": dictionary["id"],
            "infohash": dictionary["infoHash"],
            "is_best": dictionary["isBest"],
            "release_group": dictionary["releaseGroup"],
            "tracker": dictionary["tracker"],
            "updated_at": dictionary["updated"],
            "url": dictionary["url"],
        }
        return cls.model_validate(kwargs)

    @field_validator("infohash")
    @classmethod
    def _replace_placeholder_infohash(cls, value: str) -> str | None:
        """
        SeaDex API uses `<redacted>` to indicate that the torrent has no infohash (because it's private).
        This replaces it with None for a more pythonic approach.
        """
        if value.strip().casefold() == "<redacted>":  # Private torrents do not have an infohash
            return None
        return value

    @field_validator("files")
    @classmethod
    def _sort_files(cls, value: tuple[File, ...]) -> tuple[File, ...]:
        """Sort the files."""
        return tuple(natsorted(value, key=lambda file: file.name, alg=ns.PATH))


class EntryRecord(FrozenBaseModel):
    """Represents a single anime entry in SeaDex."""

    anilist_id: int
    """The AniList ID of the anime."""
    collection_id: str
    """The ID of the collection the entry belongs to."""
    collection_name: str
    """The name of the collection the entry belongs to."""
    comparisons: tuple[str, ...]
    """A tuple of comparison urls."""
    created_at: UTCDateTime
    """The timestamp of when the entry was created."""
    id: str
    """The ID of the entry."""
    is_incomplete: bool
    """Whether the entry is considered incomplete."""
    notes: str
    """Additional notes about the entry."""
    theoretical_best: str | None
    """The theoretical best release for the entry, if known."""
    torrents: tuple[TorrentRecord, ...]
    """A tuple of `TorrentRecord` objects associated with the entry."""
    updated_at: UTCDateTime
    """The timestamp of when the entry was last updated."""

    @property
    def url(self) -> str:
        """The URL of the entry."""
        return f"https://releases.moe/{self.anilist_id}/"

    @classmethod
    def _from_dict(cls, dictionary: dict[str, Any], /) -> Self:
        """Parse the response from the SeaDex API into a `EntryRecord` object."""
        kwargs = {
            "anilist_id": dictionary["alID"],
            "collection_id": dictionary["collectionId"],
            "collection_name": dictionary["collectionName"],
            "comparisons": (i.strip() for i in dictionary["comparison"].split(",") if i != ""),
            "created_at": dictionary["created"],
            "id": dictionary["id"],
            "is_incomplete": dictionary["incomplete"],
            "notes": dictionary["notes"],
            "theoretical_best": dictionary["theoreticalBest"],
            "updated_at": dictionary["updated"],
            "torrents": [TorrentRecord._from_dict(tr) for tr in dictionary["expand"]["trs"]],
        }
        return cls.model_validate(kwargs)

    @field_validator("theoretical_best")
    @classmethod
    def _replace_placeholder_infohash(cls, value: str) -> str | None:
        """
        SeaDex API uses an empty string to indicate an empty theoreticalBest field.
        This replaces it with None for a more pythonic approach.
        """
        return value if value else None



from __future__ import annotations

import base64

from stringenum import CaseInsensitiveStrEnum


class Tracker(CaseInsensitiveStrEnum):
    """Enum of public and private trackers."""

    # Public Trackers
    NYAA = "Nyaa"
    ANIMETOSHO = "AnimeTosho"
    ANIDEX = "AniDex"
    RUTRACKER = "RuTracker"
    # Private Trackers
    PRIVATE_TRACKER = "PT"
    BEYONDHD = "BeyondHD"
    PASSTHEPOPCORN = "PassThePopcorn"
    BROADCASTTHENET = "BroadcastTheNet"
    HDBITS = "HDBits"
    BLUTOPIA = "Blutopia"
    AITHER = "Aither"
    OTHER = "Other"

    def is_private(self) -> bool:
        """
        Check if the current tracker is private.

        Returns
        -------
        bool
            `True` if the tracker is private, `False` otherwise.

        """
        return False if self.value in ("Nyaa", "AnimeTosho", "AniDex", "RuTracker") else True

    def is_public(self) -> bool:
        """
        Check if the current tracker is public.

        Returns
        -------
        bool
            `True` if the tracker is public, `False` otherwise.

        """
        return not self.is_private()

    @property
    def domain(self) -> str:
        """
        Returns the domain name associated with the current tracker.

        Returns
        -------
        str
            Domain name of the tracker, or an empty string for "Other".

        """
        _map = {
            # Public Trackers
            "NYAA": b"bnlhYS5zaQ==",
            "ANIMETOSHO": b"YW5pbWV0b3Noby5vcmc=",
            "ANIDEX": b"YW5pZGV4LmluZm8=",
            "RUTRACKER": b"cnV0cmFja2VyLm9yZw==",
            # Private Trackers
            "PRIVATE_TRACKER": b"YW5pbWVieXRlcy50dg==",
            "BEYONDHD": b"YmV5b25kLWhkLm1l",
            "PASSTHEPOPCORN": b"cGFzc3RoZXBvcGNvcm4ubWU=",
            "BROADCASTTHENET": b"YnJvYWRjYXN0aGUubmV0",
            "HDBITS": b"aGRiaXRzLm9yZw==",
            "BLUTOPIA": b"Ymx1dG9waWEuY2M=",
            "AITHER": b"YWl0aGVyLmNj",
            "OTHER": b"",
        }

        return base64.b64decode(_map[self.name]).decode()
